package flare.core
{
	import flare.materials.*;
	import flare.system.*;
	import flash.events.*;
	import flash.geom.*;
	
	/**
	 * @author Ariel Nehmad
	 */
	public class Font3D extends Pivot3D implements IDrawable
	{
		public static const ALIGN_LEFT:uint 	= 	1 << 0;
		public static const ALIGN_RIGHT:uint 	= 	1 << 1;
		public static const ALIGN_CENTER_X:uint = 	1 << 2;
		public static const ALIGN_CENTER_Y:uint = 	1 << 3;
		
		private var _descriptor:XML;
		private var _charset:Charset;
		private var _charsContainer:Pivot3D;
		private var _chars:Vector.<Mesh3D> = new Vector.<Mesh3D>( 256, true );
		private var _transform:Matrix3D = new Matrix3D;
		
		public var text:String;
		public var align:uint = ALIGN_LEFT;
		
		public function Font3D( charsContainer:Pivot3D, descriptor:XML )
		{
			super( name );
		
			_descriptor = descriptor;
			_charset = new Charset( descriptor );
			_charsContainer = charsContainer;
			
			if ( _charsContainer is ILibraryExternalItem )
				ILibraryExternalItem( _charsContainer ).load();
			
			if ( _charsContainer.children.length )
				initChars();
			else
				_charsContainer.addEventListener( Event.COMPLETE, initChars );
		}
		
		private function initChars( e:Event = null ):void
		{
			_charsContainer.removeEventListener( Event.COMPLETE, initChars );
			
			for each ( var p:Pivot3D in _charsContainer.children ) {
				var id:uint = uint(p.name);
				if ( id < 256 )
					_chars[id] = p as Mesh3D;
			}
		}
		
		public function getTextWidth( text:String, startAt:int = 0, singleLine:Boolean = false ):Number
		{
			var cursorX:Number = 0;
			var maxWidth:Number = 0;
			var length:int = text.length;
			
			for ( var i:int = startAt; i < length; i++ ) {
				var id:int = text.charCodeAt(i);
				var ch:CharDesc = _charset.chars[id];
				if ( id == 10 ) {
					if ( cursorX > maxWidth ) maxWidth = cursorX;
					if ( singleLine ) return maxWidth;
					cursorX = _charset.outline;
				}
				if ( !ch ) continue;
				cursorX += ch.xAdvance;
			}
			
			if ( cursorX > maxWidth ) maxWidth = cursorX;
			return maxWidth;
		}
		
		public function getTextHeight( string:String ):Number
		{
			return getNumLines( string ) * _charset.lineHeight;
		}
		
		public function getNumLines( string:String ):uint
		{
			var count:int = 1;
			var index:int = string.indexOf( "\n", index );
			while ( index != -1 ) {
				index = string.indexOf( "\n", index + 1 );
				count++;
			}
			return count;
		}
		
		public function drawText( text:String, align:uint = 1, material:Material3D = null ):void
		{
			var cursorX:Number = 0;
			var cursorY:Number = -_charset.lineHeight * 0.75;
			var length:int = text.length;
			var index:int = 0;
			var count:int = 0;
			var i:int = 0;
			
			if ( align & ALIGN_CENTER_X ) 
				cursorX -= getTextWidth( text, 0, true ) * 0.5;
			else if ( align & ALIGN_RIGHT ) 
				cursorX -= getTextWidth( text, 0, true );
			
			if ( align & ALIGN_CENTER_Y ) 
				cursorY += getNumLines( text ) * _charset.lineHeight * 0.5;
				
			while ( i < length ) 
			{
				var id:int = text.charCodeAt(i);
				var ch:CharDesc = _charset.chars[id];
				var chPiv:Mesh3D = _chars[id];
				
				// if new line.
				if ( id == 10 ) {
					cursorX = 0;
					cursorY -= _charset.lineHeight;
					if ( align & ALIGN_CENTER_X ) 
						cursorX -= getTextWidth( text, i + 1, true ) * 0.5;
					else if ( align & ALIGN_RIGHT ) 
						cursorX -= getTextWidth( text, i + 1, true );
				}
				
				// setup char position and draw.
				if ( ch ) {
					if ( chPiv ) {
						chPiv.transform.identity();
						chPiv.x = cursorX + ch.xOffset - chPiv.bounds.min.x;
						chPiv.y = cursorY; // <- we dont need to offset here, 3ds max script already export the pivot in the correct place.
						chPiv.transform.append( world );
						chPiv.draw( false, material );
					}
					cursorX += ch.xAdvance;
				}
				
				i++;
				count++;
			}			
		}
		
		/* INTERFACE flare.core.IDrawable */
		
		override public function draw( includeChildren:Boolean = true, material:Material3D = null ):void
		{
			if ( text )
				drawText( text, align, material );
		}
	}
}



import flash.geom.*;

class CharDesc
{
	public var id:int;
	public var x:Number;
	public var y:Number;
	public var width:Number;
	public var height:Number;
	public var xOffset:Number;
	public var yOffset:Number;
	public var xAdvance:Number;
	public var page:uint;
	
	public function CharDesc( xml:XML )
	{
		this.id = xml.@id;
		this.x = xml.@x;
		this.y = xml.@y;
		this.width = xml.@width;
		this.height = xml.@height;
		this.xOffset = xml.@xoffset;
		this.yOffset = xml.@yoffset;
		this.xAdvance = xml.@xadvance;
		this.page = xml.@page;
	}
}

class Charset
{
	public var lineHeight:Number;
	public var base:Number;
	public var width:Number;
	public var height:Number;
	public var pages:uint;
	public var outline:Number;
	public var padding:Rectangle;
	public var chars:Vector.<CharDesc> = new Vector.<CharDesc>(256, true);
	
	public function Charset( xml:XML )
	{
		this.lineHeight = xml.common.@lineHeight;
		this.base = xml.common.@base;
		this.width = xml.common.@scaleW;
		this.height = xml.common.@scaleH;
		this.pages  = xml.common.@pages;
		this.outline = xml.info.@outline;
		
		var p:Array = xml.info.@padding.toString().split( "," );
		this.padding = new Rectangle();
		this.padding.top = p[0];
		this.padding.left = p[1];
		this.padding.bottom = p[2];
		this.padding.right = p[3];
		
		for each ( var x:XML in xml.chars.children() )
			chars[uint(x.@id)] = new CharDesc( x );
	}
}
